;
; Minesweeper v2.2 for Ash 3.0+, CrASH 1.1+ 
; by Wouter Demuynck (kadem@unicall.be) 
; http://uc2.unicall.be/kadem/ti82 (TI-82 ASM Corner)
;
; Revisions :
;  TI-82 Ash Minesweeper 2.2 - August 12, 1998
;   -> optimizations
;   -> scrolling method slightly changed
;   -> Ports to TI-83 SOS and Ashell
;
; Last Update : August 12, 1998
;

#include "ti82.h"       ; include ash aliases
#include "keys.inc"     ; key scancodes

;TEXT_ZT(Row,Col,Addr) places a zero terminated string at Row,Col (normal style)
#define TEXT_ZT(Row,Col,Addr) .db %00100001 \ .db Row,Col \ LD (CURSOR_POS),hl \ .db %00100001 \ .dw Addr \ ROM_CALL(D_ZT_STR)

;TEXT_ZT(X,Y,Addr) places a zero terminated string at X,Y (menu style)
#define TEXT_ZM(Xloc,Yloc,Addr) .db %00100001 \ .db Xloc,Yloc \ LD (CURSOR_X),hl \ .db %00100001 \ .dw Addr \ ROM_CALL(D_ZM_STR)

.org START_ADDR
.db "Minesweeper 2.2 by WD",0

; allocate memory
; first some things to store in the text mem

;NAME           | PLACE IN MEM     | BYTES | DESCRIPTION
;---------------------------------------------------------------------------
YVALUE          = TEXT_MEM         ; 1     | YVALUE of cursor (0->11)
XVALUE          = TEXT_MEM + 1     ; 1     | XVALUE of cursor (0->7)
TOPLEFTY        = TEXT_MEM + 2     ; 1     | Y value of top left case
TOPLEFTX        = TEXT_MEM + 3     ; 1     | X value of top left case
FLAGSPUTOUT     = TEXT_MEM + 4     ; 1     | Nr. of flags used   
NUMBER_OF_BOMBS = TEXT_MEM + 5     ; 1     | Number of bombs on the field
WIDTH           = TEXT_MEM + 6     ; 1     | width of playfield
HEIGHT          = TEXT_MEM + 7     ; 1     | height of playfield             
MAXBOMBS        = TEXT_MEM + 8     ; 1     | Max. Number Of Bombs            
MAXWIDTH        = TEXT_MEM + 9     ; 1     | Max. Width                      
MAXHEIGHT       = TEXT_MEM + 10    ; 1     | Max. Height                     
MINBOMBS        = TEXT_MEM + 11    ; 1     | Min. Number Of Bombs            
MINWIDTH        = TEXT_MEM + 12    ; 1     | Min. Width                      
MINHEIGHT       = TEXT_MEM + 13    ; 1     | Min. Height                     
CURRENTSETTING  = TEXT_MEM + 14    ; 1     | Current Setting (0-1-2) 
STATUS          = TEXT_MEM + 15    ; 1     | 1 if game over or won, else 0  
PLAYFIELD       = APD_BUF          ; 96/625| playfield(96 min/625 max)
INT_STATE       = $8D72            ; 1     | Crash Stat of Interrupts
				   ;       | (only when CrASH is on the calc)
CR_KHAND        = $8D91            ; /     | CrASH's KEY_HAND
				   ;       | (only when CrASH is on the calc)
CR_GRBCopy      = $8D94            ; /     | CrASH's Fast copy to LCD
				   ;       | (only when CrASH is on the calc)                                   

; To make customization possible, PLAYFIELD is placed in the APD_BUF
; There is a routine implemented to check if the APD_BUF is used for a
; CrASH interrupt. If an Ash interrupt is in the APD_BUF, the calc will
; most likely crash (it uses 625 bytes of the APD_BUF so there isn't much
; free space left).

; every byte in (PLAYFIELD) has the following format :
;   4 LSB's : number from 0-8 (bombs around) / 9 if it is a bomb
;   Bit 5   : set if revealed 
;   Bit 6   : set if question mark on top
;   Bit 7   : set if flag on top (cannot be revealed)

; set rom page to 4 (needed for FIND_PIXEL)
 ld   a,$8c     
 OUT  (2),A                                               

; initiate settings (copy from StartSettings to TEXT_MEM + 5):
; 20 bombs, 12 width (min), 7 height (min)
; max width = 25, max height = 25, max bombs = 255, min bombs = 10
 ld   hl,StartSettings
 ld   de,TEXT_MEM+5
 ld   bc,10
 ldir

;CHECK IF A CRASH INTERRUPT IS RUNNING (and if that's the case, stop)

  ld a,($8D4E)           ; we use this to see if CrASH is running
  cp $21
  jr nz,StartScreen      ; if crash is not running, start game
  ld a,(INT_STATE)       ; else, check if an CrASH interrupt is running
  or a                   ; (= cp 0)
  jr z,StartScreen       ; if that's not case, start the game
  ;If there is a CrASH interrupt in the mem, stop the execution of the
  ;program. I don't check if there's another 625 bytes free in the
  ;APD_BUF since no interrupt will be that small I guess (143 bytes)
  call CR_GRBCopy        ; clear LCD (CrASH is running/GRAPH_MEM is empty)
  TEXT_ZT(0,0,MsgInterrupt)   ; Show Message "Please Disable Interrupts"
  JP CR_KHAND            ; Jump to CrASH's KEY HAND (CrASH is running!)
			 ; after you press a key, you will quit to CrASH 

StartScreen:
;***************************
; Show Title
;***************************
 call CLRGBU         ; clear graph mem

 ld de,GRAPH_MEM + 3 + 60 ; displays to x= 3*8 = 24 / y = 5
 ld hl,TitlePic         ; load address of TitlePic into hl
    ;;;call PutImage    ; display title pic (PutImage by Jimmy Mardell)
    ;;;; PutImage:      ; Copies picture in (hl) to address de in GRAPH_MEM
	  ld bc,$0708   ; Changed to save some bytes
pi_newrow:              ; see Ash Online School Lesson 5 for normal version 
	  push  bc
	  push  de
pi_newcol:
	  ld  a,(hl)      
	  ld  (de),a      
	  inc hl          
	  inc de
	  djnz  pi_newcol
	  pop de
	  ld  bc,12
	  ex  de,hl
	  add hl,bc
	  ex  de,hl
	  pop bc
	  dec c
	  jr  nz,pi_newrow
	 ;;;; ret

  ROM_CALL(DISP_GRAPH)  ; copy the graph mem to the screen

  TEXT_ZM(40,13,MsgAuthor)     ; display my name : "by W.D."
  TEXT_ZM(32,7*8-5,MsgUse2nd)  ; display message : "Use 2nd to start"
  TEXT_ZT(3,0,MsgStartScreen)  ; display String : "Bombs....Width...Height"

  call ShowSettings   ; write settings on the screen

mainloop:             ; screen where you choose the ammount of bombs
 call GET_KEY         ; wait for a keypress/returns scancode to 'A' register
 cp G_MODE
 ret z                ; if MODE is pressed, quit the game
 cp G_CLEAR
 ret z                ; if CLEAR is pressed, quit the game
 cp G_2nd
 jp z,StartGame       ; if 2nd is pressed, Start the game
 cp G_ENTER
 jr z,StartGame       ; if Enter is pressed, Start the game
 cp G_RIGHT      
 jr z,MoreSettings    ; if Right arrow is pressed, Update Settings
 cp G_LEFT       
 jr z,LessSettings    ; if Left arrow is pressed, Update Settings
 cp G_UP
 jr z,UpSettings      ; if Up arrow is pressed, Update Settings
 cp G_DOWN
 jr z,DownSettings    ; if Down arrow is pressed, Update Settings
 cp G_YEDIT+1
 jr c,LoadPresets     ; if 'A' <= G_YEDIT then jump there !
 jr mainloop

LoadPresets:
 cp G_GRAPH                 ; cp b
 jr c,mainloop              ; jump there if a < G_GRAPH
 sub G_GRAPH - 1            ; 1 = G_GRAPH / 2 = TRACE / .. / 5 = YEDIT
 ; calculate the address of where the presets are
 ld hl,Presets-3
 ld b,a
 ld de,3
LoadPresetsLoop:
 add hl,de
 djnz LoadPresetsLoop
 ld b,d \ ld c,e                ;3 bytes to copy (DE is still 3!)
 ld de,NUMBER_OF_BOMBS          ;destination address
 ldir                           ;Copy the new settings
 jr SettingsChangedOK 

MoreSettings:
 ld a,3
 call CheckMaxMin
 jr z,mainloop
 inc a
 ld (hl),a
 jr SettingsChangedOK
LessSettings:
 ld a,6
 call CheckMaxMin
 jr z,mainloop
 dec a
 ld (hl),a
SettingsChangedOK:
 call ShowSettings
 jr mainloop

CheckMaxMin:
 ld hl,NUMBER_OF_BOMBS
 ld bc,(CURRENTSETTING)
 ld b,0                   ; bc contains 0,1,2
 add hl,bc
 push hl
 ld c,a          ; 'BC' is now 3 or 6
 ld a,(hl)
 add hl,bc
 cp (hl)         ; MIN OF THAT SETTING
 pop hl
 ret

UpSettings:
 ld hl,CURRENTSETTING
 dec (hl)
 ld a,-1
 cp (hl)
 jr nz,SettingsChangedOK
DownSettings:
 ld hl,CURRENTSETTING
 inc (hl)
 ld a,3
 cp (hl)
 jr nz,SettingsChangedOK
 jr UpSettings

StartGame:
;******************
; initialize cursor
;******************
 ld hl,$0101
 ld (YVALUE),hl       ; put XVALUE and YVALUE to (1,1)
 ld hl,$FFFF
 ld (TOPLEFTY),hl     ; put TOPLEFTY and TOPLEFTX to (-1,-1)
 xor a
 ld (STATUS),a        ; set status to 0 -> playing !
 ld a,(NUMBER_OF_BOMBS)
 ld (FLAGSPUTOUT),a
 call CreateLevel     ; calculate playfield information
 call ShowAll
 call DrawRemoveCursor

cursorloop:
; read keys & move cursor
; enter/del/2nd -> jump to RevealOne
; mode/clear -> to main screen
; stat/alpha -> place a flag or question mark

 ld hl,(YVALUE)            ; X -> H ; Y -> L  ; saves a lot of bytes later
 ex de,hl                  ; save hl somewhere
 call GET_KEY              ; scancode for last keypress to a register
 ex de,hl                  ; get hl back
 cp G_MODE
 jp z,StartScreen
 cp G_CLEAR
 jp z,StartScreen
 cp G_RIGHT      
 jr z,MoveRight
 cp G_LEFT       
 jr z,MoveLeft
 cp G_UP         
 jr z,MoveUp
 cp G_DOWN      
 jr z,MoveDown
 cp G_ENTER
 jp z,RevealOne
 cp G_2nd
 jp z,RevealOne
 cp G_ALPHA
 jr z,PutFlag
 jr cursorloop

MoveLeft:
 dec h
 ld a,h
 or a           ;! cp -1
 call z,ScrollItLeft
 jr MoveIt
MoveRight:
 inc h
 ld a,h
 cp 11          ;!cp 12
 call z,ScrollItRight
 jr MoveIt
MoveUp:
 dec l
 ld a,l
 or a           ;! cp -1                       
 call z,ScrollItUp
 jr MoveIt
MoveDown:
 inc l
 ld a,l
 cp 6           ;! cp 7        
 call z,ScrollItDown
MoveIt:
 call DrawRemoveCursor
 ld (YVALUE),hl
 call DrawRemoveCursor
 jp cursorloop

PutFlag:
 call GetRealCoords
 call GetBlockInfo
 bit 5,a                  ; already revealed ?
 jp nz,cursorloop
 bit 7,a                  ; flag on top ?
 jr nz,PF_Question
 bit 6,a                  ; Question Mark On Top ?
 jr nz,PF_Nothing
 ; else, put a flag
   ex de,hl
   ld hl,FLAGSPUTOUT 
   ld a,(hl)
   or a
   ex de,hl
   jp z,cursorloop
   ex de,hl
   dec (hl)
   ex de,hl
   call GetBlockInfo
 set 7,a        ; set bit 7 of a (MSB, top left bit) -> FLAGGED !
 jr PF_End
PF_Question:
 xor %11000000  ; -> remove flag and put question mark
   push hl               
   ld hl,FLAGSPUTOUT     
   inc (hl)              
   pop hl                
 jr PF_End
PF_Nothing:
 res 6,a ; unset bit 6 of a -> No question mark anymore
PF_End:
 call WriteBlockInfo
 call ShowAll
 call DrawRemoveCursor  
 jr CheckIfWon

RevealOne:           ; reveal the contents of a spot on the map
 call GetRealCoords

   ; if already revealed or question mark/flag on top, don't go on
   call GetBlockInfo   ;     -> Blockinfo in A
   and %11100000
   or a    ;cp 0 
   jp nz,cursorloop

 call CalculateAddress
 ld a,(de)    
 set 5,a         ;     -> set revealed on (bit 5)
 ld (de),a    

 and %00001111       ;     -> mask out 4 MSB's (gives us Nr. of bombs around
		     ;        or a 9 when it IS a bomb)
 cp %00001001        ; check if we have to do with a bomb
 jr z,GameOver       ; if so, game over !

;if not, we have no bomb !
 or a    ;cp 0 
 jp z,RevealSpecial ; if no bombs around, we need to call a special routine

 ; else, we can just start redrawing the playfield
EndRevealed:
 call ShowAll
 call DrawRemoveCursor

;NEW************
CheckIfWon:
 ld a,(FLAGSPUTOUT)
 or a
 jp nz,cursorloop      ; if not all bombs found yet, don't check

 ld a,(WIDTH)
 ld b,a
CIF_LoopX:            
 push bc              
 dec b                
 ld h,b               
   ld a,(HEIGHT)      
   ld b,a                    
CIF_LoopY:
   push bc            
   dec b
   ld l,b

   call CalculateAddress
   ld a,(de)

   bit 7,a
   jr z,CIF_LoopY_Cont  ; if no flag, continu routine

   ; so we have a flag, now check if it is on a bomb
   cp %10001001         ; code for flagged bomb (unrevealed of course)
   jr nz,CIF_STOP       ; if a flag placed where no bomb, STOP THE THING !

CIF_LoopY_Cont:

   pop bc           
   djnz CIF_LoopY
 pop bc             
 djnz CIF_LoopX
 jr PlayerHasWon        ;player has won when all flags are on bombs !!

CIF_STOP:    
 pop bc         
 pop bc         
 jp cursorloop

GameOver:
 ; display game over message on new screen & wait for keypress
 ld hl,STATUS
 inc (hl)
 call ShowAll
 ROM_CALL(DISP_GRAPH)
 TEXT_ZM(0,$39,MsgGameOver)

 ld b,3
GameOverShockLoop:
 push bc
 ld a,$42
 out ($10),a
 call DISP_DELAY
 call Pausetje
 ld a,$40
 out ($10),a
 call DISP_DELAY
 call Pausetje
 pop bc
 djnz GameOverShockLoop

GameOver_Loop:
 call GET_KEY
 cp G_MODE
 jp z,StartScreen
 cp G_CLEAR
 jp z,StartScreen
 cp G_2nd
 jp z,StartScreen
  push af
 cp G_UP
 call z,ScrollItUp
  pop af
 cp G_DOWN
  push af
 call z,ScrollItDown
  pop af
 cp G_LEFT
  push af
 call z,ScrollItLeft
  pop af
 cp G_RIGHT
  push af
 call z,ScrollItRight
  pop af
 jr GameOver_Loop

Pausetje:                  ; this creates a little pause 
 ld b,$09                  ; this pause is used in the
PausetjeLoop:              ; gameoverschockloop
 halt
 djnz PausetjeLoop
 ret

PlayerHasWon:
 ; display winner message on new screen & wait for keypress
 ld hl,STATUS
 inc (hl)         ; put status to 1 -> game over
 inc (hl)         ; put status to 2 -> WON ! 
 TEXT_ZM(0,$39,MsgWon)
 jr GameOver_Loop  ; wait for key 

RevealSpecial:         ; number of bombs should be in a (and should be 0)
 TEXT_ZM(0,$39,MsgWait)
 call GetRealCoords

;-----------LETS GO THROUGH THE PLAYFIELD AND START REVEALING
 ld b,2 ; let's do the whole thing 2 times (better performance)
	; 3 should be even better, but slows down the gaming considerably
RS_LoopSeveralTimes:
 push bc            ;*

 ; first time : let X go from WIDTH -> 0
 ld a,(WIDTH)
 ld b,a
RS_Loop1:
 push bc            ;*

 dec b
 ld h,b                        ; h contains X
   ld a,(HEIGHT)
   ld b,a
RS_Loop2:
   push bc              ;*   
   dec b
   ld l,b                        ; b contains Y
   call CalculateAddress   ; these two lines do the same as call Getblockinfo
   ld a,(de)               ; the address of the needed block is in de
		      ; this shoudl speed it all a little bit up (less calls)
   and %00001111
   or a  ; cp 0
   call z,RevealAllAroundIfZero
RS_Loop2_Part2:
   pop bc               ;*   
   djnz RS_Loop2

 pop bc             ;*   
 djnz RS_Loop1

 ; encore une foi 
 ; this time, let X go from 0 -> WIDTH (other direction)
 ld a,(WIDTH)
 ld b,a
RS_Loop3:
 push bc             
 ld a,(WIDTH)
 sub b        ; a = a - b = WIDTH - b ( 0..1..2..3..WIDTH-2..WIDTH-1)
 ld h,a                      ; h contains X
   ld a,(HEIGHT)
   ld b,a
RS_Loop4:
   push bc           
   ld a,(HEIGHT)
   sub b        ; a = a - b = HEIGHT - b  (0..1..2..HEIGHT-2...HEIGHT-1)
   ld l,a                        ; h contains X
   call CalculateAddress   ; these two lines do the same as call Getblockinfo
   ld a,(de)               ; the address of the needed block is in de
		      ; this shoudl speed it all a little bit up (less calls)
   and %00001111
   or a  ; cp 0
   call z,RevealAllAroundIfZero
RS_Loop4_Part2:
   pop bc            
   djnz RS_Loop4
 pop bc              
 djnz RS_Loop3

 pop bc            
 djnz RS_LoopSeveralTimes

 jp EndRevealed

RevealAllAroundIfZero:
   ld a,(de)               ; 'a' was changed, get is back, address still in de
			   ; no call -> faster
   bit 5,a
   ret z                   ; if not revealed yet, don't try
   dec l                                 
   dec h
   call RevealedOn                        ; 1             1 2 3
   inc h                                  ;               5 X 4
   call RevealedOn                        ; 2             6 7 8
   inc h
   call RevealedOn                        ; 3
   inc l
   call RevealedOn                        ; 4
   dec h
   dec h
   call RevealedOn                        ; 5
   inc l
   call RevealedOn                        ; 6
   inc h
   call RevealedOn                        ; 7
   inc h
   call RevealedOn                        ; 8
   dec h
   dec l
RS_NextPiece:
   ret

RevealedOn:
 ld a,h              
 ld bc,(WIDTH)
 cp -1               
 ret z
 cp c
 ret z
 ld a,l
 cp -1
 ret z
 cp b
 ret z
 call CalculateAddress
 ld a,(de)
 bit 7,a
 ret nz     ; don't reveal if a flag on top..
 bit 6,a
 ret nz     ; don't reveal if a question mark on top..
 set 5,a
 ld (de),a            
 ret

GetBlockInfo:
; purpose : Gets the information on a certain spot of the playfield
; input : h = X value (0-> WIDTH - 1)
;         l = Y value (0-> HEIGHT - 1)
; returns : block info in a
 call CalculateAddress
 ld a,(de)
 ret  

CalculateAddress:
; purpose : calculate the memoryaddress of a certain block of the playfield
; input : h = X value (0-> WIDTH - 1)
;         l = Y value (0-> HEIGHT - 1)
; returns : address of blockinfo in de
 push hl
 push bc

 ld a,l                        ; a contains y value (0->HEIGHT-1)
 ld b,a                        ; b contains y value (0->HEIGHT-1)

 push hl                       ; push hl to save x and y value for later

 ld hl,PLAYFIELD               ; address of playfield in hl
 or a    ;cp 0                           ; cp a (y value) to 0
 jr z,CalcAddr_GoOn            ; if y = 0 then goto CalcAddr_GoOn ( la Basic)

 ;put the WIDTH into 'de'
 push af
 ld a,(WIDTH)                  ; de = width
 ld e,a
 ld d,0
 pop af

CalcAddr_Loop:                 ; loop b times (= y value)
 add hl,de                     ; increase hl with WIDTH (1 row down)
 djnz CalcAddr_Loop

CalcAddr_GoOn:

 pop de                        ; get x and y values back from stack
			       ; but in 'de' this time
			       ; d = x (0->WIDTH-1)
			       ; e = y (0->HEIGHT-1)
 ld e,d                        ; e contains x value
 ld d,0                        ; d contains 0 -> de contains x value

 add hl,de                     ; increase hl with de (x value)
 ex de,hl                      ; exchange de and hl

 pop bc
 pop hl
 ret

WriteBlockInfo:
; purpose : Write the information to a certain spot of the playfield
; input : h = X value (0->WIDTH - 1)
;         l = Y value (0->HEIGHT - 1)
;         a = block info
; output : none
 push af
 call CalculateAddress
 pop af
 ld (de),a
 ret

DrawRemoveCursor: ; draws or removes the cursor
; Purpose : Drawing or Removing the cursor
; Input   : X(0-11) and Y(0-7) into XVALUE and YVALUE 
; Output  : Cursor drawed or removed (+ GRAPH_MEM copied to screen)
 push hl
 push af
 push bc
 ld bc,(YVALUE)  ; X->B Y->C
 ;calculate first location
 ld hl,GRAPH_MEM - 96
 ld d,0
 ld e,b
 add hl,de
 ld de,8*12
 inc c
 ld b,c
DR_CalcStart:
 add hl,de
 djnz DR_CalcStart
 ;OK, address is now in [HL]
 ld b,8
DR_Loop:
 ld a,(hl)
 xor $FF
 ld (hl),a
 ld de,12
 add hl,de    
 djnz DR_Loop
 ROM_CALL(DISP_GRAPH)
 pop bc
 pop af
 pop hl
 ret

; purpose : Creates a level
; input   : WIDTH/HEIGHT/NUMBER_OF_BOMBS in text_mem
; output  : PLAYFIELD contains level data
;*; 6/02 !!!!!! NEEdS To bE MaDE FAsTER !!!!!!!!!!!
CreateLevel:

 ; First, calculate boardsize
   ld hl,(WIDTH)
   ROM_CALL($1F18-$1A)      ; hl = h*l
   ;+;ld hl,0
   ;+;ld a,(WIDTH)
   ;+;ld b,a
   ;+;ld de,(HEIGHT)
   ;+;ld d,0  ; de contains height
;+;calcloop:
   ;+;add hl,de
   ;+;djnz calcloop
 ; 'hl' now contains the boardsize

 ; *****************************************
 ; clean PLAYFIELD !!! (Fill it with zero's)
 ; *****************************************
    ld b,h \ ld c,l        ; boardsize now loaded in bc
    ;*; push hl \ pop bc
    ld hl,PLAYFIELD
    ld de,PLAYFIELD + 1
    ld (hl),0
    ldir

 ; *****************************************
 ; Show playfield and message to wait
 ; *****************************************
 call ShowAll
 ROM_CALL(DISP_GRAPH)
 ld hl,$3900
 ld (CURSOR_X),hl  ; (0,37)
 ld hl,MsgWait
 ROM_CALL(D_ZM_STR)

 ; ******************************************************
 ; now, put the bombs in the first bytes of the playfield
 ; ******************************************************
   ;+;ld a,(NUMBER_OF_BOMBS)        ; NOTE : MAX 256 !
   ld hl,PLAYFIELD                    
   ;+;ld b,a                             
   ld bc,(NUMBER_OF_BOMBS-1)        ; Nr Of bombs -> b
   ld a,%00001001  ;  (=9 -> BOMB !)  
CL_Loop:
   ld (hl),a             
   inc hl                
   djnz CL_Loop          

 ; *******************************************************
 ; Now let's shuffle the playfield so the bombs get spread
 ; *******************************************************

 ;ld b,10
Sh_Loop:
 ;push bc
 ;+;ld a,(WIDTH)           ; WIDTH 
 ;+;ld b,a
 ld bc,(WIDTH-1)           ; b= width
Sh_L1:
 push bc
   dec b
   ld h,b               ; X
   ;+;ld a,(HEIGHT)        ; HEIGHT 
   ;+;ld b,a
   ld bc,(HEIGHT-1)
Sh_L2:
   push bc
    dec b
    ld l,b                 ; Y
    call CalculateAddress 
    ld a,(de)             
    ld b,a                 ; blockinfo 1 in b
    push hl
      ld a,(WIDTH)
      call Random
      ld h,a
	ld a,(HEIGHT)
	call Random
	ld l,a
    call CalculateAddress 
    ld a,(de) 
      ld c,a                 ; blockinfo 2 in c
      ld a,b                 ; blockinfo 1 in a
    ld (de),a
      pop hl                 ; coords 1 in hl
    ld a,c                 ; blockinfo 2 in a
    call WriteBlockInfo    ; write blockinfo 2 on location 1
   pop bc
   djnz Sh_L2
 pop bc
 djnz Sh_L1
 ;pop bc
 ;djnz Sh_Loop
 
 ; ***************************************************
 ; NOW, CALCULATE THE NUMBER OF BOMBS AROUND EACH SPOT
 ; ***************************************************
 ;+;ld a,(WIDTH)
 ;+;ld b,a
 ld bc,(WIDTH-1)
Calc_Bombs_Loop1:
 push bc
 dec b
 ld h,b                         ; h contains X
   ;+;ld a,(HEIGHT)
   ;+;ld b,a
   ld bc,(HEIGHT-1)
Calc_Bombs_Loop2:
   push bc
   dec b
   ld l,b                        ; b contains Y
   call CalculateAddress  
   ld a,(de)     
   and %00001111
   cp 9                  ; don't try if we have a bomb (duh)
   jr z,NextPiece                        ;  1 2 3
   push hl                               ;  4 X 6    X = place where
   dec l                                 ;  7 8 9        we check
   ld b,0                ; bomb counter to zero
   dec h                                     
   call CheckBomb                        ; 1
   inc l
   call CheckBomb                        ; 4
   inc l
   call CheckBomb                        ; 7
   inc h
   dec l
   dec l
   call CheckBomb                        ; 2
   inc l
   inc l
   call CheckBomb                        ; 8
   inc h
   dec l
   dec l
   call CheckBomb                        ; 3
   inc l
   call CheckBomb                        ; 6
   inc l
   call CheckBomb                        ; 9
   dec l
   dec l
   pop hl
   ld a,b         ; number of bombs around -> a (use this as blockinfo)
   call WriteBlockInfo
NextPiece:
   pop bc
   djnz Calc_Bombs_Loop2
 pop bc
 djnz Calc_Bombs_Loop1
 ret        ; end of createlevel (at last)

CheckBomb:
 ; first, check if h and l (x and y) aren't out of the playfield
 ld a,h        ; if that's the problem, return

 cp -1
 ret z
   ex de,hl        ; save hl in de
   ld hl,WIDTH
 cp (hl)
   ex de,hl        ; get hl back
 ret z
 ld a,l
 cp -1
 ret z
   ex de,hl
   ld hl,HEIGHT
 cp (hl)
   ex de,hl
 ret z
 call CalculateAddress 
 ld a,(de) 
 and %00001111
 cp 9
 ret nz
 inc b       ; only if we have a bomb (we use b as the bomb counter)
 ret

; purpose : Shows the settings on the main screen
; input   : CURRENTSETTING/WIDTH/HEIGHT/NUMBER_OF_BOMBS
; output  : DATA on screen
ShowSettings:
 ld hl,$0103    

 ld (CURSOR_POS),hl
 ld a,' '
 call MINES_TX_CHARPUT 

 inc l
 ld (CURSOR_POS),hl
 call MINES_TX_CHARPUT 

 inc l
 ld (CURSOR_POS),hl
 call MINES_TX_CHARPUT 

 ld hl,$0103

 ld a,(CURRENTSETTING)
 add a,l
 ld l,a
 ld (CURSOR_POS),hl  ; (X=14,Y=3)
 ld a,5                                ; show pointer
 call MINES_TX_CHARPUT 
   ld hl,$0D03       
   ld (CURSOR_POS),hl  ; (X=14,Y=3)
   ld a,(NUMBER_OF_BOMBS)
   call DisplayNumberInA
 ld hl,$0D04         
 ld (CURSOR_POS),hl  ; (X=14,Y=4)
 ld a,(WIDTH)
 call DisplayNumberInA
   ld hl,$0D05       
   ld (CURSOR_POS),hl  ; (X=14,Y=5)
   ld a,(HEIGHT)
   call DisplayNumberInA
 ret

MINES_TX_CHARPUT:
 ROM_CALL(TX_CHARPUT)
 ret

; purpose : Displays the number stored in 'A' on the screen (normal text)
; input   : A
;           CURSOR_POS contain most right cursor position
; output  : number on screen
DisplayNumberInA:
 ld hl,0
 ld l,a
 ld b,3
DispNrInALoop:
 CALL UNPACK_HL
 add a,48                 ; add 49 to get the ASCII code
 call MINES_TX_CHARPUT 
 push hl
 ld hl,CURSOR_COL      
 dec (hl)              
 dec (hl)              
 pop hl
 djnz DispNrInALoop
 ret
 
; purpose : Draws all sprites on the screen
; input   : STATUS,PLAYFIELD
; output  : screen updated
ShowAll:
 ;+;ld a,(STATUS)        ; 2 for won, 1 for game over, 0 for normal
 ;+;ld c,a
 ld bc,(STATUS)          ; status -> c

 push hl           ; save original hl
 ld hl,(YVALUE)
 push hl           ; YVALUE WILL CHANGE CONSTANTLY DURING THIS ROUTINE
		   ; BETTER SAVE IT

 push bc
 call CLRGBU      ; clear the graph memory
 pop bc

 ld hl,(TOPLEFTY)        ; topleft y -> l ; topleftx -> h

 ld b,12                 ; 12 blocks from left to right
SA_LoopX:
 push bc
 push hl

 dec b
 ld d,b      ; X SCREEN VALUE IN D
 ld a,b
 add a,h
 ld h,a                  ; h contains ABSOLUTE X VALUE !!!

   ld b,7 ;**                  ; 7 block from top to bottom
SA_LoopY:
   push de
   push bc
   push hl

   dec b
   ld e,b                ; Y value of cursor in e
     ex de,hl
     ld (YVALUE),hl      ; write screen cursor coords to YVALUE
     ex de,hl
   ld a,b
   add a,l
   ld l,a                ; 'h' contains ABSOLUTE Y VALUE !!!

   ; Show border if necessary
   ld a,-1
   cp l
   jr z,SA_Border
   cp h
   jr z,SA_Border
   ld a,(WIDTH)
   cp h
   jr z,SA_Border
   ld a,(HEIGHT)
   cp l
   jr z,SA_Border
   
   call GetBlockInfo     ; returns blockinfo in 'a'

     ;start checking
     bit 7,a              
     jr nz,SA_Flag        ; check if a flag needs to be placed
     bit 6,a              
     jr nz,SA_Question    ; check if a question mark needs to be placed
     bit 5,a              
     jr z,SA_Ground       ; if not revealed yet, display ground sprite
     ; if none of above, it is a number (or the revealed bomb)
     and %00001111
SA_ShowIt:
     call DispNumber
     ;end of checking
SA_Next:

   pop hl
   pop bc
   pop de

   djnz SA_LoopY

 pop hl
 pop bc

 djnz SA_LoopX

 jr SA_End
SA_Flag:
   ld a,(STATUS)
   or a
   jr nz,SpecialGameOverTreatment
NormalTreatment:
 ld a,10                ; 10 will show flag sprite
 jr SA_ShowIt
SpecialGameOverTreatment:
   call GetBlockInfo 
   and %00001111     
   cp 9
   jr z,NormalTreatment
   ld a,13                ; 13 will show 'wrong' sprite
   jr SA_ShowIt

SA_Question:
 call CheckIfBomb
 jr z,SA_Bomb2Show
 ld a,11                ; 11 will show Question sprite
 jr SA_ShowIt

SA_Border: 
 ld a,14                ; 11 will show Question sprite
 jr SA_ShowIt

SA_Ground:
 call CheckIfBomb
 jr z,SA_Bomb2Show
 ld a,12                ; 12 will show ground sprite
 jr SA_ShowIt

SA_Bomb2Show:
 ld a,c
 or a   
 jp z,SA_ShowIt   ; shows the ground sprite if unrevealed bombs aren't shown
 ld a,9                 ; 9 will show bomb
 jp SA_ShowIt

SA_End:
 ; draw line under playfield
 ld c,7
 ld de,$005F
 call DrawHorLine
 ld c,$50
 ld de,$0007
 call DrawVerLine
 ld c,$3A
 call DrawVerLine


 ld hl,(TOPLEFTY)       ; h = X coord of upper left corner
			; l = Y coord of upper left corner
 ld a,h
    push hl
 cp -1 ;!or a  
  push af
 call nz,DrawLeftSprite   ; if we can scroll left, draw indicator
  pop af
 add a,12
 ld bc,(WIDTH)          ; width in c , height in b
 inc c ;!
 cp c
 call nz,DrawRightSprite  ; if we can scroll right, draw indicator
    pop hl
 ld a,l
 cp -1 ;!or a  
   push af
 call nz,DrawUpSprite  ; if we can scroll up, draw indicator
   pop af
 ld bc,(WIDTH)
 inc b ;!
 add a,7        
 cp b
 call nz,DrawDownSprite  ; if we can scroll down, draw indicator

 ld a,9                  ; (bomb)
 ld hl,$0907             ; X = 9 Y = 7 
 ld (YVALUE),hl
 call DispNumber

     
   set 7,(IY+$14)     ; write menu text to graph mem only
   ld hl,$3945
   ld (CURSOR_X),hl
   ld hl,0
   ld a,(FLAGSPUTOUT)
   ld l,a
   ld b,3
DispNrInAMenuTextLoop:
   CALL UNPACK_HL
   add a,48                 ; 49 toevoegen om ASCII code te krijgen
   ROM_CALL(M_CHARPUT)     ; tonen van karakter
   ld a,(CURSOR_X)  
   sub 8
   ld (CURSOR_X),a
   djnz DispNrInAMenuTextLoop
   res 7,(IY+$14)         ; write menu text to screen only


 pop hl                ; get original YVALUE BACK
 ld (YVALUE),hl
 pop hl                ; get original 'hl' back

 ret

CheckIfBomb:
 ld a,c        ; first check if bombs are shown of not
 cp 2        
 jr z,EndCheckIfBomb_NoBomb ; if won, don't show bombs
 or a    
 jr z,EndCheckIfBomb_NoBomb ; if just playing, don't show them !
 call GetBlockInfo ; if we are in normal mode, we need to check
 and %00001111     ; for a bomb and we need to draw it if there is one
EndCheckIfBomb_NoBomb:
 cp 9              ; if bombs are not shown this will unset the z flag
 ret               ; else z flag will be set if it is a bomb

DispNumber:
 push bc
 push de
 ld b,a
 ld hl,Small0
 or a 
 jr z,DispNumber_2
 ld de,6     
DispNumber_1:
 add hl,de
 djnz DispNumber_1
DispNumber_2:
 call DispSprite
 pop de
 pop bc
 ret

; purpose : Displays a sprite on the screen (using putsprite)
; input   : address in HL (X and Y in XVALUE and YVALUE)
; output  : sprite drawn in GRAPH_MEM
DispSprite:       
 push bc
 push hl
 ld hl,(YVALUE)   
 add hl,hl      ; h * 2 / l * 2
 add hl,hl      ; h * 4 / l * 4
 add hl,hl      ; h * 8 / l * 8
 ld b,h \ ld c,l     ; ld bc,hl
 pop hl
 inc b
 inc c
 ld de,$0606       ; 6 x 6 sprite      
 call PutSprite
 pop bc
 ret

DrawRightSprite:
 ld hl,RightSprite
 ld bc,$5C3A
 ld de,$0305 
 jr DrawThatSprite

DrawLeftSprite:
 ld hl,LeftSprite
 ld bc,$523A
 ld de,$0305 
 jr DrawThatSprite

DrawUpSprite:
 ld hl,UpSprite
 ld bc,$5639
 ld de,$0503 
 jr DrawThatSprite

DrawDownSprite:
 ld hl,DownSprite
 ld bc,$563D
 ld de,$0503 

DrawThatSprite:
 call PutSprite
 ret


; purpose : Draws a vertical line
; input   : c = x; d = y1 < e = y2   (0,0) is coord of lower left corner !!
; output  : line drawn in GRAPH_MEM
DrawVerLine:    
 ld a,e
 sub d
 ld b,a                   ; loop e - d = y2 - y1 = delta y times
DrawVerLineLoop:
 push bc
 ld a,e
 sub b
 push bc ;* push bc
 ld c,a
 pop hl  ; get it back in HL 
 ld b,l
 push de
 call PutPixel
 pop de
 pop bc
 djnz DrawVerLineLoop
 ret

; purpose : Draws a horizontal line
; input   : c = y; d = x1 < e = x2   (0,0) is coord of lower left corner !!
; output  : line drawn in GRAPH_MEM
DrawHorLine:      
 ld a,e
 sub d
 ld b,a
DrawHorLineLoop:
 push bc
 push de
 ld a,e
 sub b
 ld b,a
 call PutPixel
 pop de
 pop bc
 djnz DrawHorLineLoop
 ret


; purpose : Calculate the absolute coordinates
; input   : XVALUE/YVALUE -> cursor position on screen 
;           TOPLEFTX/TOPLEFTY -> coords of upper left corner visible surface
; output  : Abs coords in h(x) and l(y)
GetRealCoords:
 ld bc,(YVALUE)
 ld hl,(TOPLEFTY)
 ld a,h
 add a,b
 ld h,a
 ld a,l
 add a,c
 ld l,a
 ret

ScrollItLeft:
 inc h
 ld a,(TOPLEFTX)
 cp -1  ;! or a   
 ret z
 dec a
 ld (TOPLEFTX),a
 jr ScrollItNow
ScrollItRight:
 dec h
 ld a,(TOPLEFTX)
 add a,12
 ld bc,(WIDTH) ; width in c
 inc c  ;!
 cp c
 ret z
 sub 12
 inc a
 ld (TOPLEFTX),a
 jr ScrollItNow
ScrollItUp:
 inc l
 ld a,(TOPLEFTY)
 cp -1 ;! or a    ;cp 0 
 ret z
 dec a
 ld (TOPLEFTY),a
 jr ScrollItNow
ScrollItDown:
 dec l
 ld a,(TOPLEFTY)
 add a,7
 ld bc,(WIDTH) ; height in b
 inc b   ;!
 cp b
 ret z
 sub 7
 inc a
 ld (TOPLEFTY),a
ScrollItNow:
 call ShowAll
 ld a,(STATUS)
 or a ; cp 0      (but 'A' is destroyed)
 call z,DrawRemoveCursor ; if not game over or won, draw cursor
 ret z
 ROM_CALL(DISP_GRAPH)
 ret

; -----------------------------------------------------------------------
; | START OF ROUTINES NOT WRITTEN BY ME                                 |
; -----------------------------------------------------------------------

; purpose : Puts a pixel at B,C
; input   : B,C contain X and Y (0,0) is lower left corner
; output  : Pixel is put on
; writer  : Jimmy Mardell
PutPixel: 
 CALL FIND_PIXEL          
 ld de,GRAPH_MEM
 add hl,de
 or (hl)
 ld (hl),a
 ret

; purpose : Clear the graph mem
; input   : /
; output  : GRAPH_MEM is cleared
CLRGBU:
 ld     hl,GRAPH_MEM
 ld     de,GRAPH_MEM+1
 ld   bc,767 
 ld     (hl),0
 ldir
 ret

; purpose : Puts a sprite stored at (HL) at B,C  
; input   : HL -> address of the sprite
;           B,C -> X and Y coords (0,0) is upper left corner of screen
;           D,E -> Width and height of sprite
; output  : GRAPH_MEM is cleared
; writer  : Jimmy Mardell (slightly addapted by Wouter Demuynck)
PutSprite: 
 push   bc 
 ld   a,63 
 sub    c
 ld     c,a
 push   hl
 push   hl
 call   $4166            ;find pixel
 ld b,d         
 ld c,e         
 ld     de,$88B8
 add    hl,de
 ex     de,hl
 pop    hl
 push   hl
 pop    ix
 ex     de,hl
PS_NewRow:
 push   bc
 ld     d,(ix)
 inc    ix
 push   af
 push   hl
PS_NewCol:
 rl     d
 ld     e,a
 jr     nc,PS_NoPixel
 or     (hl)
;* ld   (hl),a
 jr     PS_NextPixel
PS_NoPixel:
 cpl
 and    (hl)
;* ld   (hl),a
PS_NextPixel:
 ld   (hl),a ;*
 ld     a,e
 rrca
 jr     nc,PS_SameByte
 inc    hl
PS_SameByte:
 djnz   PS_NewCol
 pop    hl
 pop    af
 ld     de,12
 add    hl,de
 pop    bc
 dec    c
 jr     nz,PS_NewRow
 pop    hl
 pop    bc
 ret

; purpose : Puts a sprite stored at (HL) at B,C  
; input   : A     (0 <= x < A) with x the random number to get
; output  : A is random number
; writer  : Jimmy Mardell 
Random:
  push  bc     
	push    de
	push    hl
	ld      b,a
	ld      a,r
	add     a,a
	ld      hl,0
	ld      d,0
	ld      e,a
RMul:
	add     hl,de
	djnz    RMul
	ld      a,h
	pop     hl
	pop     de
	pop     bc
  ret       

; -----------------------------------------------------------------------
; |    PROGRAM DATA (SPRITES/PICTURES/TEXT)                             |
; -----------------------------------------------------------------------

TitlePic: ; MINES title
;.db 7,8 ; width(in bytes) and height(in bits)
.db %01111111,%11100001,%10001111,%11100000,%11111110,%00011111,%11100000
.db %11111111,%11110001,%10001111,%11110001,%11111110,%00111111,%11100000
.db %11000000,%00110000,%00001100,%00110001,%10000000,%00110000,%00000000
.db %11000110,%00110001,%10001100,%00110001,%10011100,%00111111,%11000000
.db %11000110,%00110001,%10001100,%00110001,%10011100,%00011111,%11100000
.db %11000110,%00110001,%10001100,%00110001,%10000000,%00000000,%01100000
.db %11000110,%00110001,%10001100,%00110001,%11111110,%00111111,%11100000
.db %11000000,%00110001,%10001100,%00110000,%11111110,%00111111,%11000000

Small0: ;(6,6 sprite)
.db %00000000
.db %00000000
.db %00000000
.db %00000000
.db %00000000
.db %00000000
Small1: ;(6,6 sprite)
.db %00100000
.db %01100000
.db %00100000
.db %00100000
.db %00100000
.db %01110000
Small2: ;(6,6 sprite)
.db %01110000
.db %10001000
.db %00001000
.db %00110000
.db %01000000
.db %11111000
Small3: ;(6,6 sprite)
.db %11111000
.db %00010000
.db %00100000
.db %00010000
.db %10001000
.db %01110000
Small4: ;(6,6 sprite)
.db %00010000
.db %00110000
.db %01010000
.db %11111000
.db %00010000
.db %00010000
Small5: ;(6,6 sprite)
.db %11111000
.db %10000000
.db %11110000
.db %00001000
.db %00001000
.db %11110000
Small6: ;(6,6 sprite)
.db %01111000
.db %10000000
.db %11110000
.db %10001000
.db %10001000
.db %01110000
Small7: ;(6,6 sprite)
.db %11111000
.db %00001000
.db %00010000
.db %00100000
.db %01000000
.db %01000000
Small8: ;(6,6 sprite)
.db %01110000
.db %10001000
.db %01110000
.db %10001000
.db %10001000
.db %01110000
SmallBomb:  ; (6*6 sprite)            ; 9
.db %00011000
.db %00100000
.db %01110000
.db %11111000
.db %11111000
.db %01110000
SmallFlag:   ; (6,6 sprite)           ; 10
.db %00100100
.db %01111100
.db %11011100
.db %10000100
.db %00000100
.db %00000100 
SmallQuestion: ;(6,6 sprite)          ; 11
.db %01110000
.db %10001000
.db %00010000
.db %00100000
.db %00000000
.db %00100000
GroundSprite: ;(6,6 sprite)           ; 12
.db %10000000
.db %00010000
.db %00000000
.db %01000000
.db %00000000
.db %00000100
WrongSprite: ;(6,6 sprite)           ; 13
.db %10000100
.db %01001000
.db %00110000
.db %00110000
.db %01001000
.db %10000100
BorderSprite: ;(6,6 sprite)           ; 14
.db %00000000
.db %00000000
.db %00110000
.db %00110000
.db %00000000
.db %00000000


LeftSprite:
.db %00100000
.db %01100000
.db %11100000
.db %01100000
.db %00100000
RightSprite:
.db %10000000
.db %11000000
.db %11100000
.db %11000000
.db %10000000
UpSprite:
.db %00100000
.db %01110000
.db %11111000
DownSprite:
.db %11111000
.db %01110000
.db %00100000

StartSettings:
.db 20,12,7,255,25,25,10,12,7,0
Presets:     ; presets from difficult to easy
;.db NR_OF_BOMBS, WIDTH, HEIGHT
.db 255,25,25                     ; 7.5
.db 99,25,20                      ; 4.9
.db 40,16,16                      ; 3.6
.db 25,13,14                      ; 2.7
.db 10,12,7                       ; 1.6
				  ; difficulty (0-10)
				  ; calculated : 10-((width*height)/bombs)
MsgAuthor:
.db "by WD",0
MsgWon:
.db "You Win !",0
MsgGameOver:
.db "Game Over...",0
MsgStartScreen:
.db "  Bombs  :      "
.db "  Width  :      "
.db "  Height :",0
MsgUse2nd:
.db "2nd to start",0
MsgWait:
.db "Please Wait...",0
MsgInterrupt:
.db "Please disable  "
.db "Interrupts first",0

.end

